# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np


class IStandardKernel:
    """Interface for a standard kernel k(x, x') that in principle can be integrated.

    .. seealso::
       * :class:`emukit.quadrature.interfaces.IRBF`
       * :class:`emukit.quadrature.interfaces.IProductMatern52`
       * :class:`emukit.quadrature.interfaces.IProductMatern32`
       * :class:`emukit.quadrature.interfaces.IProductBrownian`

    """

    def K(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        """The kernel k(x1, x2) evaluated at x1 and x2.

        :param x1: First argument of the kernel, shape (n_points N, input_dim)
        :param x2: Second argument of the kernel, shape (n_points M, input_dim)
        :returns: Kernel evaluated at x1, x2, shape (N, M).
        """
        raise NotImplementedError

    # the following methods are gradients of the kernel
    def dK_dx1(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        """Gradient of the kernel wrt x1 evaluated at pair x1, x2.

        :param x1: First argument of the kernel, shape (n_points N, input_dim)
        :param x2: Second argument of the kernel, shape (n_points M, input_dim)
        :return: The gradient of the kernel wrt x1 evaluated at (x1, x2), shape (input_dim, N, M)
        """
        raise NotImplementedError

    def dKdiag_dx(self, x: np.ndarray) -> np.ndarray:
        """The gradient of the diagonal of the kernel (the variance) v(x):=k(x, x) evaluated at x.

        :param x: The locations where the gradient is evaluated, shape (n_points, input_dim).
        :return: The gradient of the diagonal of the kernel evaluated at x, shape (input_dim, n_points).
        """
        raise NotImplementedError


class IRBF(IStandardKernel):
    r"""Interface for an RBF kernel.

    .. math::
        k(x, x') = \sigma^2 e^{-\frac{1}{2}\sum_{i=1}^{d}r_i^2},

    where :math:`d` is the input dimensionality,
    :math:`r_i = \frac{x_i-x_i'}{\lambda_i}` is the scaled vector difference of dimension :math:`i`,
    :math:`\lambda_i` is the :math:`i` th element of the ``lengthscales`` property
    and :math:`\sigma^2` is the ``variance`` property.

    .. note::
        Inherit from this class to wrap your standard RBF kernel. The wrapped kernel can then be
        handed to a quadrature RBF kernel that augments it with integrability.

    .. seealso::
       * :class:`emukit.quadrature.kernels.QuadratureRBF`
       * :class:`emukit.quadrature.kernels.QuadratureRBFLebesgueMeasure`
       * :class:`emukit.quadrature.kernels.QuadratureRBFGaussianMeasure`

    """

    @property
    def lengthscales(self) -> np.ndarray:
        r"""The lengthscales :math:`\lambda` of the kernel."""
        raise NotImplementedError

    @property
    def variance(self) -> float:
        r"""The scale :math:`\sigma^2` of the kernel."""
        raise NotImplementedError

    def dK_dx1(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        scaled_vector_diff = np.swapaxes((x1[None, :, :] - x2[:, None, :]) / self.lengthscales**2, 0, -1)
        return -self.K(x1, x2)[None, ...] * scaled_vector_diff

    def dKdiag_dx(self, x: np.ndarray) -> np.ndarray:
        return np.zeros((x.shape[1], x.shape[0]))


class IProductMatern12(IStandardKernel):
    r"""Interface for a Matern12 (a.k.a. Exponential) product kernel.

    The product kernel is of the form
    :math:`k(x, x') = \sigma^2 \prod_{i=1}^d k_i(x, x')` where

    .. math::
        k_i(x, x') = e^{-r_i}.

    :math:`d` is the input dimensionality,
    :math:`r_i:=\frac{|x_i - x'_i|}{\lambda_i}`,
    :math:`\sigma^2` is the ``variance`` property and :math:`\lambda_i` is the :math:`i` th element
    of the ``lengthscales`` property.

    Make sure to encode only a single variance parameter, and not one for each individual :math:`k_i`.

    .. note::
        Inherit from this class to wrap your standard product Matern12 kernel. The wrapped kernel can then be
        handed to a quadrature product Matern12 kernel that augments it with integrability.

    .. seealso::
       * :class:`emukit.quadrature.kernels.QuadratureProductMatern12`
       * :class:`emukit.quadrature.kernels.QuadratureProductMatern12LebesgueMeasure`

    """

    @property
    def nu(self) -> float:
        """The smoothness parameter of the kernel."""
        return 0.5

    @property
    def lengthscales(self) -> np.ndarray:
        r"""The lengthscales :math:`\lambda` of the kernel."""
        raise NotImplementedError

    @property
    def variance(self) -> float:
        r"""The scale :math:`\sigma^2` of the kernel."""
        raise NotImplementedError

    def _dK_dx1_1d(self, x1: np.ndarray, x2: np.ndarray, lengthscale: float) -> np.ndarray:
        """Unscaled gradient of 1D Matern12 where ``lengthscale`` is the lengthscale parameter.

        This method can be used in case the product Matern12 is implemented via a List of
        univariate Matern12 kernels.

        :param x1: First argument of the kernel, shape = (n_points N,).
        :param x2: Second argument of the kernel, shape = (n_points M,).
        :param lengthscale: The lengthscale of the 1D Matern12.
        :return: The gradient of the kernel wrt x1 evaluated at (x1, x2), shape (N, M).
        """
        r = (x1.T[:, None] - x2.T[None, :]) / lengthscale  # N x M
        dr_dx1 = r / (lengthscale * abs(r))
        dK_dr = -np.exp(-abs(r))
        return dK_dr * dr_dx1

    def dKdiag_dx(self, x: np.ndarray) -> np.ndarray:
        return np.zeros((x.shape[1], x.shape[0]))


class IProductMatern32(IStandardKernel):
    r"""Interface for a Matern32 product kernel.

    The product kernel is of the form
    :math:`k(x, x') = \sigma^2 \prod_{i=1}^d k_i(x, x')` where

    .. math::
        k_i(x, x') = (1 + \sqrt{3}r_i ) e^{-\sqrt{3} r_i}.

    :math:`d` is the input dimensionality,
    :math:`r_i:=\frac{|x_i - x'_i|}{\lambda_i}`,
    :math:`\sigma^2` is the ``variance`` property and :math:`\lambda_i` is the :math:`i` th element
    of the ``lengthscales`` property.

    Make sure to encode only a single variance parameter, and not one for each individual :math:`k_i`.

    .. note::
        Inherit from this class to wrap your standard product Matern32 kernel. The wrapped kernel can then be
        handed to a quadrature product Matern32 kernel that augments it with integrability.

    .. seealso::
       * :class:`emukit.quadrature.kernels.QuadratureProductMatern32`
       * :class:`emukit.quadrature.kernels.QuadratureProductMatern32LebesgueMeasure`

    """

    @property
    def nu(self) -> float:
        """The smoothness parameter of the kernel."""
        return 1.5

    @property
    def lengthscales(self) -> np.ndarray:
        r"""The lengthscales :math:`\lambda` of the kernel."""
        raise NotImplementedError

    @property
    def variance(self) -> float:
        r"""The scale :math:`\sigma^2` of the kernel."""
        raise NotImplementedError

    def _dK_dx1_1d(self, x1: np.ndarray, x2: np.ndarray, lengthscale: float) -> np.ndarray:
        """Unscaled gradient of 1D Matern32 where ``lengthscale`` is the lengthscale parameter.

        This method can be used in case the product Matern32 is implemented via a List of
        univariate Matern32 kernels.

        :param x1: First argument of the kernel, shape = (n_points N,).
        :param x2: Second argument of the kernel, shape = (n_points M,).
        :param lengthscale: The lengthscale of the 1D Matern32.
        :return: The gradient of the kernel wrt x1 evaluated at (x1, x2), shape (N, M).
        """
        r = (x1.T[:, None] - x2.T[None, :]) / lengthscale  # N x M
        dr_dx1 = r / (lengthscale * abs(r))
        dK_dr = -3 * abs(r) * np.exp(-np.sqrt(3) * abs(r))
        return dK_dr * dr_dx1

    def dKdiag_dx(self, x: np.ndarray) -> np.ndarray:
        return np.zeros((x.shape[1], x.shape[0]))


class IProductMatern52(IStandardKernel):
    r"""Interface for a Matern52 product kernel.

    The product kernel is of the form
    :math:`k(x, x') = \sigma^2 \prod_{i=1}^d k_i(x, x')` where

    .. math::
        k_i(x, x') = (1 + \sqrt{5} r_i + \frac{5}{3} r_i^2) \exp(- \sqrt{5} r_i).

    :math:`d` is the input dimensionality,
    :math:`r_i:=\frac{|x_i - x'_i|}{\lambda_i}`,
    :math:`\sigma^2` is the ``variance`` property and :math:`\lambda_i` is the :math:`i` th element
    of the ``lengthscales`` property.

    Make sure to encode only a single variance parameter, and not one for each individual :math:`k_i`.

    .. note::
        Inherit from this class to wrap your standard product Matern52 kernel. The wrapped kernel can then be
        handed to a quadrature product Matern52 kernel that augments it with integrability.

    .. seealso::
       * :class:`emukit.quadrature.kernels.QuadratureProductMatern52`
       * :class:`emukit.quadrature.kernels.QuadratureProductMatern52LebesgueMeasure`

    """

    @property
    def nu(self) -> float:
        """The smoothness parameter of the kernel."""
        return 2.5

    @property
    def lengthscales(self) -> np.ndarray:
        r"""The lengthscales :math:`\lambda` of the kernel."""
        raise NotImplementedError

    @property
    def variance(self) -> float:
        r"""The scale :math:`\sigma^2` of the kernel."""
        raise NotImplementedError

    def _dK_dx1_1d(self, x1: np.ndarray, x2: np.ndarray, lengthscale: float) -> np.ndarray:
        """Unscaled gradient of 1D Matern52 where ``lengthscale`` is the lengthscale parameter.

        This method can be used in case the product Matern52 is implemented via a List of
        univariate Matern52 kernels.

        :param x1: First argument of the kernel, shape = (n_points N,).
        :param x2: Second argument of the kernel, shape = (n_points M,).
        :param lengthscale: The lengthscale of the 1D Matern52.
        :return: The gradient of the kernel wrt x1 evaluated at (x1, x2), shape (N, M).
        """
        r = (x1.T[:, None] - x2.T[None, :]) / lengthscale  # N x M
        dr_dx1 = r / (lengthscale * abs(r))
        dK_dr = (-5 / 3) * np.exp(-np.sqrt(5) * abs(r)) * (abs(r) + np.sqrt(5) * r**2)
        return dK_dr * dr_dx1

    def dKdiag_dx(self, x: np.ndarray) -> np.ndarray:
        return np.zeros((x.shape[1], x.shape[0]))


class IBrownian(IStandardKernel):
    r"""Interface for a Brownian motion kernel.

    .. math::
        k(x, x') = \sigma^2 \operatorname{min}(x, x')\quad\text{with}\quad x, x' \geq 0,

    where :math:`\sigma^2` is the ``variance`` property.

    .. note::
        Inherit from this class to wrap your standard Brownian motion kernel.
        The wrapped kernel can then be handed to a quadrature Brownian motion kernel that
        augments it with integrability.

    .. seealso::
       * :class:`emukit.quadrature.kernels.QuadratureBrownian`
       * :class:`emukit.quadrature.kernels.QuadratureBrownianLebesgueMeasure`

    """

    @property
    def variance(self) -> float:
        r"""The scale :math:`\sigma^2` of the kernel."""
        raise NotImplementedError

    def dK_dx1(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        x1_rep = np.repeat(x1[:, 0][np.newaxis, ...], x2.shape[0], axis=0).T
        x2_rep = np.repeat(x2[:, 0][np.newaxis, ...], x1.shape[0], axis=0)
        return self.variance * (x1_rep < x2_rep)[np.newaxis, :, :]

    def dKdiag_dx(self, x: np.ndarray) -> np.ndarray:
        return self.variance * np.ones((x.shape[1], x.shape[0]))


class IProductBrownian(IStandardKernel):
    r"""Interface for a Brownian product kernel.

    The product kernel is of the form
    :math:`k(x, x') = \sigma^2 \prod_{i=1}^d k_i(x, x')` where

    .. math::
        k_i(x, x') = \operatorname{min}(x_i-c, x_i'-c)\quad\text{with}\quad x_i, x_i' \geq c,

    :math:`d` is the input dimensionality,
    :math:`\sigma^2` is the ``variance`` property
    and :math:`c` is the ``offset`` property.

    Make sure to encode only a single variance parameter, and not one for each individual :math:`k_i`.

    .. note::
        Inherit from this class to wrap your standard product Brownian kernel. The wrapped kernel can then be
        handed to a quadrature product Brownian kernel that augments it with integrability.

    .. seealso::
       * :class:`emukit.quadrature.kernels.QuadratureProductBrownian`
       * :class:`emukit.quadrature.kernels.QuadratureProductBrownianLebesgueMeasure`

    """

    @property
    def variance(self) -> float:
        r"""The scale :math:`\sigma^2` of the kernel."""
        raise NotImplementedError

    @property
    def offset(self) -> float:
        r"""The offset :math:`c` of the kernel."""
        raise NotImplementedError

    def _dK_dx1_1d(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        """Unscaled gradient of 1D Brownian kernel.

        This method can be used in case the product Brownian is implemented via a List of
        Brownian kernels.

        :param x1: First argument of the kernel, shape = (n_points N,).
        :param x2: Second argument of the kernel, shape = (n_points M,).
        :return: The gradient of the kernel wrt x1 evaluated at (x1, x2), shape (N, M).
        """
        x1_rep = np.repeat(x1[np.newaxis, ...], x2.shape[0], axis=0).T
        x2_rep = np.repeat(x2[np.newaxis, ...], x1.shape[0], axis=0)
        return x1_rep < x2_rep



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Optional

import GPy
import numpy as np

from emukit.model_wrappers.gpy_quadrature_wrappers import create_emukit_model_from_gpy_model
from emukit.quadrature.loop import VanillaBayesianQuadratureLoop
from emukit.quadrature.measures import IntegrationMeasure
from emukit.quadrature.methods.vanilla_bq import VanillaBayesianQuadrature
from emukit.quadrature.typing import BoundsType


def create_vanilla_bq_loop_with_rbf_kernel(
    X: np.ndarray,
    Y: np.ndarray,
    integral_bounds: Optional[BoundsType] = None,
    measure: Optional[IntegrationMeasure] = None,
    rbf_lengthscale: float = 1.0,
    rbf_variance: float = 1.0,
) -> VanillaBayesianQuadratureLoop:
    """Creates a quadrature loop with a standard (vanilla) model.

    :param X: Initial training point locations, shape (n_points, input_dim).
    :param Y: Initial training point function values, shape (n_points, 1).
    :param integral_bounds: List of d tuples, where d is the dimensionality of the integral and the tuples contain the
                            lower and upper bounds of the integral
                            i.e., [(lb_1, ub_1), (lb_2, ub_2), ..., (lb_d, ub_d)].
                            Only used if ``measure`` is not given in which case the unnormalized Lebesgue measure is used.
    :param measure: An integration measure. Either ``measure`` or ``integral_bounds`` must be given.
                    If both ``integral_bounds`` and ``measure`` are given, ``integral_bounds`` is disregarded.
    :param rbf_lengthscale: The lengthscale of the rbf kernel, defaults to 1.
    :param rbf_variance: The variance of the rbf kernel, defaults to 1.
    :return: The vanilla BQ loop.

    """

    if measure is not None and measure.input_dim != X.shape[1]:
        raise ValueError(
            f"Dimensionality of measure ({measure.input_dim}) does not match the dimensionality of "
            f"the data ({X.shape[1]})."
        )

    if (integral_bounds is not None) and (len(integral_bounds) != X.shape[1]):
        raise ValueError(
            f"Dimension of integral bounds ({len(integral_bounds)}) does not match the input dimension "
            f"of X ({X.shape[1]})."
        )

    if rbf_lengthscale <= 0:
        raise ValueError(f"rbf lengthscale must be positive. The current value is {rbf_lengthscale}.")
    if rbf_variance <= 0:
        raise ValueError(f"rbf variance must be positive. The current value is {rbf_variance}.")

    gpy_model = GPy.models.GPRegression(
        X=X, Y=Y, kernel=GPy.kern.RBF(input_dim=X.shape[1], lengthscale=rbf_lengthscale, variance=rbf_variance)
    )

    # This function handles the omittion if the integral bounds in case measure is also given.
    emukit_model = create_emukit_model_from_gpy_model(
        gpy_model=gpy_model, integral_bounds=integral_bounds, measure=measure
    )
    emukit_method = VanillaBayesianQuadrature(base_gp=emukit_model, X=X, Y=Y)
    emukit_loop = VanillaBayesianQuadratureLoop(model=emukit_method)
    return emukit_loop



import evalset.test_funcs
import GPy
from pbbo import *

# Select one of the following as the inference method: MCMCComparisonGP, EPComparisonGP, VIComparisonGPMF, VIComparisonGPFR
# We demonstrate here with the EP inference:
inference = EPComparisonGP

# Select one of the following as the acquisition function: SumOfVariances, QExpectedImprovement, ThompsonSampling
# We demonstrate here with the Thompson sampling
acquisition = ThompsonSampling

# Select objective, we demonstrate here with the 4 dimensional Sushi data
objective = evalset.test_funcs.Sushi()

# Normalize the objective and add noise:
objective.init_normalize_X()  # scale bounds between 0 and 1
objective.init_normalize_Y()  # scale function maximum and minimum between 0 and 1
noise_level = 0.05
objective = evalset.test_funcs.Noisifier(
    objective, "add", noise_level
)  # add normally distributed noise with std 'noise_level'

# Generate a kernel for the GP
kernel = GPy.kern.RBF(
    input_dim=len(objective.bounds), ARD=True, variance=1.0, lengthscale=[0.5] * len(objective.bounds)
)
kernel.variance.constrain_fixed(1.0, warning=False)
kernel.lengthscale.constrain_fixed([0.5] * len(objective.bounds), warning=False)

# Some parameters for the acquisition function
options_acquisition = {
    "acq_samples": 5000,  # How many samples are used in stochastic optimization of EV and QEI
    "acq_opt_restarts": 30,  # How many different initialization locations are randomly sampled for optimizing the acquisition function}
    "pool": -1,  # How many posterior samples are used from the posterior, if -1 all are used. Only matters for MCMC.
}

# Acquisition function optimizer options, see possible options from scipy optimize options
optimizer_options = {}


# Options for the Bayesian optimization:
config = {
    "inference": inference,
    "kernel": kernel,
    "acquisition": acquisition(options_acquisition, optimizer_options),
    "batch_size": 4,  # Batch size >2
    "initial_size": 0,  # If comparative feedback is used
    "initial_size_direct": 0,  # How many direct observations are used in initailization
    "noise": noise_level,  # Noise standard deviation for observations the likelihood uses
    "max_num_observations": 60,  # Max number of observation in the BO loo. Number of iterations is maximum number oservations divided by the batch size.
}

# Generate and run the Bayesian Optimization
bo = BayesianOptimization(config)
X, yc = bo.bayesian_optimization(objective)  # X contains the points and yc contains the comparative observations



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


"""
This module provides functions to convert from a list representation of multi-fidelity data to an array representation.

The list based representation is a list of numpy arrays, with a numpy array for every fidelity. The list is ordered
from the lowest fidelity to the highest fidelity.

The array representation is one array for all inputs where the last column of the X array is a zero-based index
indicating the fidelity.
"""
from typing import List, Tuple

import numpy as np


def convert_x_list_to_array(x_list: List) -> np.ndarray:
    """
    Converts list representation of features to array representation
    :param x_list: A list of (n_points x n_dims) numpy arrays ordered from lowest to highest fidelity
    :return: An array of all features with the zero-based fidelity index appended as the last column
    """
    # First check everything is a 2d array
    if not np.all([x.ndim == 2 for x in x_list]):
        raise ValueError("All x arrays must have 2 dimensions")

    x_array = np.concatenate(x_list, axis=0)
    indices = []
    for i, x in enumerate(x_list):
        indices.append(i * np.ones((len(x), 1)))

    x_with_index = np.concatenate((x_array, np.concatenate(indices)), axis=1)
    return x_with_index


def convert_y_list_to_array(y_list: List) -> np.ndarray:
    """
    Converts list representation of outputs to array representation
    :param y_list: A list of (n_points x n_outputs) numpy arrays representing the outputs
                   ordered from lowest to highest fidelity
    :return: An array of all outputs
    """
    if not np.all([y.ndim == 2 for y in y_list]):
        raise ValueError("All y arrays must have 2 dimensions")
    return np.concatenate(y_list, axis=0)


def convert_xy_lists_to_arrays(x_list: List, y_list: List) -> Tuple[np.ndarray, np.ndarray]:
    """
    Converts list representation of targets to array representation
    :param x_list: A list of (n_points x n_dims) numpy arrays ordered from lowest to highest fidelity
    :param y_list: A list of (n_points x n_outputs) numpy arrays representing the outputs
                   ordered from lowest to highest fidelity
    :return: Tuple of (x_array, y_array) where
             x_array contains all inputs across all fidelities with the fidelity index appended as the last column
             and y_array contains all outputs across all fidelities.
    """

    if len(x_list) != len(y_list):
        raise ValueError("Different number of fidelities between x and y")

    # Check same number of points in each fidelity
    n_points_x = np.array([x.shape[0] for x in x_list])
    n_points_y = np.array([y.shape[0] for y in y_list])
    if not np.all(n_points_x == n_points_y):
        raise ValueError("Different number of points in x and y at the same fidelity")

    return convert_x_list_to_array(x_list), convert_y_list_to_array(y_list)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import List, Tuple

import numpy as np

from emukit.core.loop.user_function import UserFunctionWrapper


def hennig2D() -> Tuple[UserFunctionWrapper, List[Tuple[float, float]]]:
    r"""2D toy integrand coined by Philipp Hennig.

    .. math::
        f(x) = e^{-x'Sx -\sin(3\|x\|^2)}

    :return: The wrapped test function, and the integrals bounds
             (the latter default to [-3, 3]^2).
    """
    integral_bounds = 2 * [(-3.0, 3.0)]
    return UserFunctionWrapper(_hennig2D), integral_bounds


def _hennig2D(x: np.ndarray, S: np.ndarray = None) -> np.ndarray:
    """
    :param x: Locations for evaluation (num_points, 2).
    :return: The function values at x, shape (num_points, 2).
    """
    if S is None:
        S = np.array([[1, 0.5], [0.5, 1]])
    f = np.exp(-np.sin(3 * np.sum(x**2, axis=1)) - np.sum((x @ S) * x, axis=1))
    return np.reshape(f, [x.shape[0], 1])



from typing import Tuple

import numpy as np
import scipy.linalg
import scipy.optimize

from emukit.core.interfaces import IModel


class SimpleGaussianProcessModel(IModel):
    """
    This model is a Gaussian process with an RBF kernel, with no ARD. It is used to demonstrate uses of emukit,
    it does not aim to be flexible, robust or fast.
    """

    def __init__(self, x: np.ndarray, y: np.ndarray):
        """
        :param x: (n_points, n_dims) array containing training features
        :param y: (n_points, 1) array containing training targets
        """
        self.x = x
        self.y = y
        self.lengthscale = 1
        self.kernel_variance = 1
        self.likelihood_variance = 1
        self.jitter = 1e-6

    def __repr__(self):
        """
        Prints the hyper-parameters
        """
        return "Lengthscale: {:.4f} \n Kernel variance: {:.4f} \n Likelihood variance: {:.4f}".format(
            self.lengthscale, self.kernel_variance, self.likelihood_variance
        )

    def optimize(self) -> None:
        """
        Optimize the three hyperparameters of the model, namely the kernel variance, kernel lengthscale and likelihood
        variance
        """

        def optimize_fcn(log_hyper_parameters):
            # take exponential to ensure positive values
            hyper_parameters = np.exp(log_hyper_parameters)
            self.lengthscale = hyper_parameters[0]
            self.kernel_variance = hyper_parameters[1]
            self.likelihood_variance = hyper_parameters[2]
            return self._negative_marginal_log_likelihood()

        lower_bound = np.log(1e-6)
        upper_bound = np.log(1e8)

        bounds = [(lower_bound, upper_bound) for _ in range(3)]
        scipy.optimize.minimize(
            optimize_fcn,
            np.log(np.array([self.lengthscale, self.kernel_variance, self.likelihood_variance])),
            bounds=bounds,
        )

    def predict(self, x_new: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Predict from model

        :param x_new: (n_points, n_dims) array containing points at which the predictive distributions will be computed
        :return: Tuple containing two (n_points, 1) arrays representing the mean and variance of the predictive
                 distribution at the specified input locations
        """
        K = self._calc_kernel(self.x)
        K += np.identity(self.x.shape[0]) * (self.jitter + self.likelihood_variance)

        L = np.linalg.cholesky(K)

        K_xs = self._calc_kernel(self.x, x_new)

        tmp = scipy.linalg.solve_triangular(L, K_xs, lower=True)
        tmp2 = scipy.linalg.solve_triangular(L, self.y, lower=True)

        mean = np.dot(tmp.T, tmp2)
        variance = (self.kernel_variance - np.sum(np.square(tmp), axis=0))[:, None]
        return mean, variance

    def _calc_kernel(self, X: np.ndarray, X2: np.ndarray = None) -> np.ndarray:
        """
        Implements an RBF kernel with no ARD

        :param X: array of shape (n_points_1, n_dims) containing input points of first argument to kernel function
        :param X2: array of shape (n_points_2, n_dims) containing input points of second argument to kernel function.
                   If not supplied K(X, X) is computed.
        :return: Kernel matrix K(X, X2) or K(X, X) if X2 not supplied.
        """
        if X2 is None:
            X2 = X

        X1sq = np.sum(np.square(X), 1)
        X2sq = np.sum(np.square(X2), 1)
        r2 = -2.0 * np.dot(X, X2.T) + (X1sq[:, None] + X2sq[None, :])
        r2 = np.clip(r2, 0, np.inf)
        return self.kernel_variance * np.exp(-0.5 * r2 / self.lengthscale**2)

    def _negative_marginal_log_likelihood(self) -> float:
        """
        :return: Negative marginal log likelihood of model with current hyper-parameters
        """
        K = self._calc_kernel(self.x)

        # Add some jitter to the diagonal
        K += np.identity(self.x.shape[0]) * (self.jitter + self.likelihood_variance)

        # cholesky decomposition of covariance matrix
        L = np.linalg.cholesky(K)

        # Log determinant of the covariance matrix
        log_det = 2.0 * np.sum(np.log(np.diag(L)))

        # calculate y^T K^{-1} y using the cholesky of K
        tmp = scipy.linalg.solve_triangular(L, self.y, lower=True)
        alpha = scipy.linalg.solve_triangular(L.T, tmp, lower=False)
        log_2_pi = np.log(2 * np.pi)
        return -0.5 * (-self.y.size * log_2_pi - self.y.shape[1] * log_det - np.sum(alpha * self.y))

    def set_data(self, X: np.ndarray, Y: np.ndarray) -> None:
        """
        Set training data to new values

        :param X: (n_points, n_dims) array containing training features
        :param Y: (n_points, 1) array containing training targets
        """
        self.x = X
        self.y = Y

    @property
    def X(self) -> np.ndarray:
        return self.x

    @property
    def Y(self) -> np.ndarray:
        return self.y



"""Integration measures."""

from .domain import BoxDomain
from .gaussian_measure import GaussianMeasure
from .integration_measure import IntegrationMeasure
from .lebesgue_measure import LebesgueMeasure

__all__ = ["BoxDomain", "IntegrationMeasure", "GaussianMeasure", "LebesgueMeasure"]



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from ...core.loop import FixedIntervalUpdater, ModelUpdater, OuterLoop
from ...core.loop.loop_state import create_loop_state
from ...core.parameter_space import ParameterSpace
from ..loop.point_calculators import BayesianMonteCarloPointCalculator
from ..methods import WarpedBayesianQuadratureModel


class BayesianMonteCarlo(OuterLoop):
    """The loop for Bayesian Monte Carlo (BMC).


    Nodes are samples from the integration measure.
    Implemented as described in Section 2.1 of the paper.

    .. rubric:: References

    C.E. Rasmussen and Z. Ghahramani, Bayesian Monte Carlo,
    Advances in Neural Information Processing Systems 15 (NeurIPS) 2003


    .. note::
        The BMC point calculator does not depend on past observations. Thus, running this BQ loop
        should be equivalent to sampling all points with MC from the measure,
        evaluating them as batch and then fitting a model to them.
        The purpose of this loop is convenience, as it can be used with the same interface
        as the active and adaptive learning schemes where point acquisition depends explicitly or implicitly
        (through hyperparameters) on the previous evaluations.

        Hint: The default :attr:`model_updater` :class:`FixedIntervalUpdater` updates and optimizes the model
        after each new sample. Since the sampling scheme of BMC does not depend on the model, alternatively,
        the dummy updater :class:`NoopModelUpdater` may be used which does not update the model.
        This may save compute time. However, the model then needs to be updated manually after the loop ran:
        i) the collected nodes are stored in ``model.loop_state``.
        ii) call ``model.set_data(loop_state.X, loop_state.Y)``
        iii) call ``model.optimize()``.

    .. seealso::
        * :class:`emukit.quadrature.loop.point_calculators.BayesianMonteCarloPointCalculator`
        * :class:`emukit.core.loop.model_updaters.NoopModelUpdater`

    :param model: A warped Bayesian quadrature model.
    :param model_updater: Defines how and when the quadrature model is updated if new data arrives,
                          defaults to :class:`FixedIntervalUpdater`.


    """

    def __init__(self, model: WarpedBayesianQuadratureModel, model_updater: ModelUpdater = None):
        if model_updater is None:
            model_updater = FixedIntervalUpdater(model, 1)

        space = ParameterSpace(model.reasonable_box_bounds.convert_to_list_of_continuous_parameters())
        candidate_point_calculator = BayesianMonteCarloPointCalculator(model, space)
        loop_state = create_loop_state(model.X, model.Y)

        super().__init__(candidate_point_calculator, model_updater, loop_state)

        self.model = model



import pickle

try:
    import torch
except ImportError:
    raise ImportError("pytorch is not installed. Please install it by running pip install torch torchvision")

from functools import partial
from typing import Tuple

from emukit.core import ContinuousParameter, ParameterSpace
from emukit.core.loop.user_function import UserFunctionWrapper
from emukit.examples.profet.meta_benchmarks.architecture import get_default_architecture
from emukit.examples.profet.meta_benchmarks.meta_surrogates import objective_function


def meta_svm(fname_objective: str, fname_cost: str, noise: bool = True) -> Tuple[UserFunctionWrapper, ParameterSpace]:
    """
    Interface to the Meta-SVM benchmark which imitates the hyperparameter optimization of a
    support vector machine on OpenML like classification datasets.
    Offline generated function samples can be download here:

    http://www.ml4aad.org/wp-content/uploads/2019/05/profet_data.tar.gz

    NOTE: make sure that the index for the objective function and the cost function match,
    e.g for sample_objective_i.pkl and sample_cost_i.pkl the index i should be the same.
    Each of those files contain the parameters of a neural network which represents a single function sample and
    some additional information, such as the mean and variance of the input normalization.

    For further information about Profet and the generated meta-surrogate benchmarks see:

    Meta-Surrogate Benchmarking for Hyperparameter Optimization
    A. Klein and Z. Dai and F. Hutter and N. Lawrence and J. Gonzalez
    arXiv:1905.12982 [cs.LG] (2019)

    :param fname_objective: filename for the objective function
    :param fname_cost: filename for the cost function
    :param noise: determines whether to add noise on the function value or not
    :return: Tuple of user function object and parameter space
    """
    parameter_space = ParameterSpace(
        [
            ContinuousParameter("log_C", 0, 1),  # scaled to [0, 1], original space was in [-10, 10]
            ContinuousParameter("log_gamma", 0, 1),
        ]
    )  # scaled to [0, 1] original space was in [-10, 10]
    data = pickle.load(open(fname_objective, "rb"))

    x_mean_objective = data["x_mean"]
    x_std_objective = data["x_std"]
    task_feature_objective = data["task_feature"]
    objective = get_default_architecture(x_mean_objective.shape[0], classification=True).float()
    objective.load_state_dict(data["state_dict"])

    data = pickle.load(open(fname_cost, "rb"))

    x_mean_cost = data["x_mean"]
    x_std_cost = data["x_std"]
    y_mean_cost = data["y_mean"]
    y_std_cost = data["y_std"]
    task_feature_cost = data["task_feature"]
    cost = get_default_architecture(x_mean_cost.shape[0]).float()
    cost.load_state_dict(data["state_dict"])

    f = partial(
        objective_function,
        model_objective=objective,
        model_cost=cost,
        task_feature_objective=task_feature_objective,
        task_feature_cost=task_feature_cost,
        x_mean_objective=x_mean_objective,
        x_std_objective=x_std_objective,
        x_mean_cost=x_mean_cost,
        x_std_cost=x_std_cost,
        y_mean_objective=None,
        y_std_objective=None,
        y_mean_cost=y_mean_cost,
        y_std_cost=y_std_cost,
        log_objective=False,
        with_noise=noise,
    )

    return f, parameter_space



import numpy as np

from emukit.core import ContinuousParameter, InformationSourceParameter, ParameterSpace
from emukit.core.loop.user_function import MultiSourceFunctionWrapper


def multi_fidelity_non_linear_sin(high_fidelity_noise_std_deviation=0, low_fidelity_noise_std_deviation=0):
    """
    Two level non-linear sin function where high fidelity is given by:

    .. math::
        f_{high}(x) = (x - \sqrt{2}) f_{low}(x)^2

    and the low fidelity is:

    .. math::
        f_{low}(x) = \sin(8 \pi x)

    Reference:
    Nonlinear information fusion algorithms for data-efficient multi-fidelity modelling.
    P. Perdikaris, M. Raissi, A. Damianou, N. D. Lawrence and G. E. Karniadakis (2017)
    http://web.mit.edu/parisp/www/assets/20160751.full.pdf
    """

    parameter_space = ParameterSpace([ContinuousParameter("x1", -5, 10), InformationSourceParameter(2)])
    user_function = MultiSourceFunctionWrapper(
        [
            lambda x: nonlinear_sin_low(x, low_fidelity_noise_std_deviation),
            lambda x: nonlinear_sin_high(x, high_fidelity_noise_std_deviation),
        ]
    )
    return user_function, parameter_space


def nonlinear_sin_low(x, sd=0):
    """
    Low fidelity version of nonlinear sin function
    """

    return np.sin(8 * np.pi * x) + np.random.randn(x.shape[0], 1) * sd


def nonlinear_sin_high(x, sd=0):
    """
    High fidelity version of nonlinear sin function
    """

    return (x - np.sqrt(2)) * nonlinear_sin_low(x, 0) ** 2 + np.random.randn(x.shape[0], 1) * sd



from typing import Callable, Dict, List, Tuple

import GPy
import numpy as np
import scipy as sp
from GPy.inference.latent_function_inference.posterior import Posterior
from GPy.util import choleskies
from GPy.util.linalg import dpotrs, dtrtrs, jitchol, pdinv, tdot
from GPy.util.univariate_Gaussian import cdfNormal, derivLogCdfNormal, logCdfNormal, std_norm_cdf, std_norm_pdf
from scipy.integrate import dblquad, quad
from scipy.linalg import block_diag, inv, sqrtm
from scipy.special import logsumexp
from scipy.stats import multivariate_normal, norm

from ..util import adam

# Helper functions
phi = lambda x: cdfNormal(x)
sigmoid = lambda x: 1.0 / (1 + np.exp(-x))

dlogphi_df = lambda x: derivLogCdfNormal(x)
dlogsigmoid_df = lambda x: sigmoid(-x)
d2logsigmoid_df = lambda x: -sigmoid(-x) * sigmoid(x)


def dL_fr(L: np.ndarray, dsigma: np.ndarray, alpha: np.ndarray, beta: np.ndarray, K: np.ndarray):
    """
    Partial derivative of function with respect to beta using cholesky decomposition and the generalized matrix chain rule
    The ending _fr comes from "full rank", this method should be used when we use the full rank parametrization

    :param L: Cholesky decomposition of Sigma: Sigma= L L^T
    :param dsigma: derivative of the function with respect to Sigma
    :param alpha: alpha parameter
    :param beta: beta parameter, the vector the derivative is taken with respect to
    :param K: prior covariance matrix
    :return: The derivative of function with respect to beta
    """
    Sigma = L @ L.T
    t2 = np.zeros_like(dsigma)
    for m in range(t2.shape[0]):
        for n in range(m + 1):
            dl = np.zeros_like(dsigma)
            dl[m, :] = L[:, n].T
            dl += dl.T
            t2[m, n] = np.trace(dsigma.T @ dl)
    return t2[np.newaxis, :]


def dL_mf(L: np.ndarray, dsigma: np.ndarray, alpha: np.ndarray, beta: np.ndarray, K: np.ndarray):
    """
    Partial derivative of function with respect to beta using cholesky decomposition and the generalized matrix chain rule
    The ending _fr comes from "mean field", this method should be used when we use the meaan field parametrization

    :param L: Cholesky decomposition of Sigma: Sigma= L L^T
    :param dsigma: derivative of the function with respect to Sigma
    :param alpha: alpha parameter
    :param beta: beta parameter, the vector the derivative is taken with respect to
    :param K: prior covariance matrix
    :return: The derivative of function with respect to beta
    """
    dL = np.zeros((L.shape[0], L.shape[1], beta.shape[0]))
    dL2 = np.zeros((L.shape[0], L.shape[1], beta.shape[0]))
    res = np.zeros_like(beta)

    S = L @ L.T
    for k in range(beta.shape[0]):
        for i in range(L.shape[0]):
            for j in range(L.shape[1]):
                dL[i, j, k] = -2.0 * beta[k] * S[i, k] * S[j, k]
        res[k] += np.trace(dsigma.T @ dL[:, :, k])
    return res


def dSigma_dLmn(L: np.ndarray, m: int, n: int):
    """
    Partial derivative of Sigma with respect to one element of L when Sigma=L L^T

    :param L: Cholesky decomposition of sigma
    :param m: row index in L
    :param n: column index in L
    :return: partial derivative of Sigma
    """
    delta = 1e-5
    L_new = L.copy()
    L_new[m, n] += delta
    Sigma_new = L_new @ L_new.T
    Sigma = L @ L.T
    return (Sigma_new - Sigma) / delta


def comp_y_ij(mu: np.ndarray, Sigma: np.ndarray, i: int, j: int, epsilon: float):
    """
    A helper method to compute the mean and covariance of y_j-y_j when we know the joint distribution of f and y=f+noise, where noise has standard deviation epsilon
    :param mu: mean of latent function f
    :param Sigma: covariance of latent function f
    :param i: index i in y_j - y_i
    :param j: index j in y_j - y_i
    :param epsilon: noise standard deviation of y
    """
    m_diff = mu[j] - mu[i]
    sigma_diff = np.sqrt(Sigma[i, i] + Sigma[j, j] + 2 * Sigma[i, j])
    return m_diff + sigma_diff * epsilon, sigma_diff


def variational_expectations_ove_full_rank(
    mf: np.ndarray, Sigma: np.ndarray, ind_winners: List[int], ind_loosers: List[int], sigma2s: float = 1.0
) -> Tuple[float, np.ndarray, np.ndarray]:
    """
    Computes the variational expectation and derivatives for the full rank approximation for a single batch

    :param mf: mean of the latent function approximation of the batch
    :param Sigma: Covariance of the latent function approximation of the batch
    :param ind_winners: List of batch winners in each pairwise comparisons. We assume that the feedback is given as batch winner form
    :param ind_loosers: List of batch loosers in each pairwise comparisons
    :param sigma2s: noise variance of the observations
    :return: expectation and its derivatives with respect to mean and covariance
    """
    N = mf.shape[0]

    dF_dm = np.zeros((N, 1))
    dF_dSigma = np.zeros((N, N))

    # Integration by quadrature
    gh_x, gh_w = np.polynomial.hermite.hermgauss(25)
    gh_w = gh_w / np.sqrt(np.pi)

    # to make sigmoid look more like probit
    sigma2s = sigma2s / 1.6

    F = 0

    # i is the winner:
    i = ind_winners[0]
    for j in ind_loosers:
        y_ij, sigma_ij = comp_y_ij(mf, Sigma, i, j, gh_x)
        F += np.sum(np.log(sigmoid(y_ij / (np.sqrt(2) * sigma2s[i]))) * gh_w)
        ms_y_ij = sigmoid(-y_ij / (np.sqrt(2) * sigma2s[i]))
        dF_dm[i, 0] += np.sum(-gh_w * ms_y_ij / (np.sqrt(2) * sigma2s[i]))
        dF_dm[j, 0] = np.sum(gh_w * ms_y_ij / (np.sqrt(2) * sigma2s[i]))
        dF_dSigma[j, j] = 0.5 * np.sum(gh_w * ms_y_ij / sigma_ij * gh_x / (np.sqrt(2) * sigma2s[i]))
        dF_dSigma[i, i] += dF_dSigma[j, j].copy()
        dF_dSigma[i, j] = 2.0 * dF_dSigma[j, j]
    return F, dF_dm, dF_dSigma


def df_d(
    y: List[Tuple[int, float]],
    yc: List[List[Tuple[int, int]]],
    m: np.ndarray,
    L: np.ndarray,
    L_inv: np.ndarray,
    K: np.ndarray,
    sigma2s: np.ndarray,
    alpha: np.ndarray,
    beta: np.ndarray,
    s_to_l: Callable = dL_fr,
):
    """
    Computes the log marginal likelihood and its derivatives with respect to alpha and beta. Works for both mean feald and full rank approximations

    :param y: Direct observations in as a list of tuples telling location index (row in X) and observation value.
    :param yc: Batch comparisons in a list of lists of tuples. Each batch is a list and tuples tell the comparisons (winner index, loser index)
    :param m: mean of the latent values
    :param L: Cholesky decomposition of the latent value covariance
    :param L_inv: inverse of the cholesky decomposition
    :param K: prior covariance
    :param sigma2s: noise variance of the observations
    :param alpha: Alpha vector used to parametrize the posterior approximation
    :param beta: Beta vector/matrix used to parametrize the posterior approximation
    :param s_to_l: A function to compute the derivative of log likelihood with respect to beta using the generalized chain rule and when we know the derivative of log likelihood with respect to Sigma
    :return: A tuple containing log marginal likelihood, its derivative with respect to alpha and its derivative with respect to beta
    """
    Sigma = L @ L.T

    dF_dm_full = np.zeros_like(m)
    dF_dSigma_full = np.zeros_like(Sigma)
    F_full = 0
    # log_marginal = 0
    d_list = np.random.choice(range(len(yc)), size=len(yc), replace=False)
    for batch_idx in d_list:
        loc_inds_winners, loc_inds_losers = [yc[batch_idx][k][0] for k in range(len(yc[batch_idx]))], [
            yc[batch_idx][k][1] for k in range(len(yc[batch_idx]))
        ]
        loc_inds_batch = np.sort(np.unique(loc_inds_winners + loc_inds_losers))
        # get winners
        ind_winners, ind_losers = [np.where(loc_inds_batch == it)[0][0] for it in loc_inds_winners], [
            np.where(loc_inds_batch == it)[0][0] for it in loc_inds_losers
        ]

        # get variational moments
        F_batch, dF_dm_batch, dF_dSigma_batch = variational_expectations_ove_full_rank(
            m[loc_inds_batch],
            Sigma[np.ix_(loc_inds_batch, loc_inds_batch)],
            ind_winners,
            ind_losers,
            sigma2s[loc_inds_batch],
        )
        dF_dm_full[loc_inds_batch] += dF_dm_batch
        dF_dSigma_full[np.ix_(loc_inds_batch, loc_inds_batch)] += dF_dSigma_batch
        F_full += F_batch

    # delta = 1e-5
    if len(y) > 0:
        ys = np.zeros((len(y), 1))
        y_inds = np.zeros(len(y), dtype=int)
        # dir_list = np.random.choice(range(len(y)), size=len(y), replace=False)
        for ind in range(len(y)):
            (y_inds[ind], ys[ind, 0]) = y[ind]  # index in kernel, y value
        F_full += -0.5 * np.sum(
            ((m[y_inds] - ys) ** 2 + Sigma[y_inds, y_inds].reshape((-1, 1))) / sigma2s[y_inds].reshape((-1, 1))
        )
        dF_dm_full[y_inds] += (ys - m[y_inds]) / sigma2s[y_inds].reshape((-1, 1))
        dF_dSigma_full[y_inds, y_inds] += -0.5 / sigma2s[y_inds].reshape((-1))

    alpha_grad = K @ dF_dm_full

    beta_grad = s_to_l(L, dF_dSigma_full, alpha, beta, K)

    log_marginal = F_full
    if beta_grad.shape[1] > 1:
        beta_grad = choleskies._triang_to_flat_pure(beta_grad)
    return log_marginal, alpha_grad, beta_grad


def recompute_posterior_fr(
    alpha: np.ndarray, beta: np.ndarray, K: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Recompute the posterior approximation (for the full rank approximation) mean: K alpha, covariance inv(K + beta)
    :param alpha: Alpha vector used to parametrize the posterior approximation
    :param beta: Beta vector/matrix used to parametrize the posterior approximation
    :param K: prior covariance
    :return: Tuple containing the mean and cholesky of the covariance, its inverse and derivatives of the KL divergence with respect to beta and alpha
    """
    N = K.shape[0]
    L = choleskies._flat_to_triang_pure(beta)
    assert L.shape[0] == 1
    L = L[0, :, :]
    lam_sqrt = np.diag(L)
    lam = lam_sqrt**2

    # Compute Mean
    m = K @ alpha
    jitter = 1e-5
    dKL_da = m.copy()
    Kinv = np.linalg.inv(K + np.eye(N) * jitter)
    L_inv = np.linalg.inv(L)

    Sigma = np.empty((alpha.size, alpha.shape[0]))
    Lamda_full_rank = np.dot(L, L.T)

    dKL_db_triang = -dL_fr(L, 0.5 * (np.linalg.inv(Lamda_full_rank) - Kinv), None, None, None)

    mat1 = np.linalg.inv(K + Lamda_full_rank)
    # Sigma = np.linalg.inv(Kinv + np.linalg.inv(Lamda_full_rank))
    Sigma = Lamda_full_rank
    # Compute KL
    KL = 0.5 * (-N + (m.T @ Kinv @ m) + np.trace(Kinv @ Sigma) - np.log(np.linalg.det(Sigma @ Kinv)))
    dKL_db = choleskies._triang_to_flat_pure(dKL_db_triang)

    return m, L, L_inv, KL, dKL_db, dKL_da


def recompute_posterior_mf(
    alpha: np.ndarray, beta: np.ndarray, K: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Recompute the posterior approximation (for the mean field approximation) mean: K alpha, covariance inv(K + beta)
    :param alpha: Alpha vector used to parametrize the posterior approximation
    :param beta: Beta vector/matrix used to parametrize the posterior approximation
    :param K: prior covariance
    :return: Tuple containing the mean and cholesky of the covariance, its inverse and derivatives of the KL divergence with respect to beta and alpha
    """
    N = alpha.shape[0]
    # Lambda = diag(lam) = diag(beta.^2)
    lam_sqrt = beta.ravel()
    lam = beta.ravel() ** 2

    # Handle A = I + Lambda*K*Lambda
    KB = K @ np.diag(lam_sqrt)
    BKB = np.diag(lam_sqrt) @ KB
    A = np.eye(N) + BKB
    Ai, LA, Li, Alogdet = pdinv(A)

    # Compute Mean
    m = K @ alpha

    # Compute covariance matrix
    W = Li @ np.diag(1.0 / lam_sqrt)  # can be accelerated using broadcasting instead of matrix multiplication
    Sigma = (
        np.diag(1.0 / lam) - W.T @ W
    )  # computes np.diag(1./lam) - np.diag(1. / lam_sqrt) @ Ai @ np.diag(1. / lam_sqrt)

    # Compute KL
    KL = 0.5 * (Alogdet + np.trace(Ai) - N + np.sum(m * alpha))

    # Compute Gradients
    A_A2 = Ai - Ai.dot(Ai)
    dKL_db = np.diag(np.dot(KB.T, A_A2)).reshape(-1, 1)
    # dKL_da = K @ alpha
    dKL_da = m.copy()

    L = GPy.util.linalg.jitchol(Sigma)
    L_inv = np.linalg.inv(L)

    return m, L, L_inv, KL, dKL_db, dKL_da


def log_lik(
    x: np.ndarray, arg_list: Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, Callable]
) -> Tuple[np.array, np.array]:
    """
    Computes the log likelihood and gradients for specific alpha and beta values concatenated in x.

    :param x: Concatenated and flattened alpha and beta
    :param arg_list: List of arguments that don't change during the optimization
                     (prior covariance, noise of the observations, observations,
                     comparisons, function to transfer the partial derivative)
    :return: Tuple containing the log marginal and its derivative with respect to alpha and beta.
    """

    K, sigma2s, y, yc, recompute_posterior, s_to_l = (
        arg_list[0],
        arg_list[1],
        arg_list[2],
        arg_list[3],
        arg_list[4],
        arg_list[5],
    )
    alpha = x[: K.shape[0]].reshape(-1, 1)
    beta = x[K.shape[0] :].reshape(-1, 1)

    if not isinstance(sigma2s, np.ndarray):
        sigma2s = sigma2s * np.ones((K.shape[0], 1))

    m, L, L_inv, KL, dKL_db, dKL_da = recompute_posterior(alpha, beta, K)

    log_marginal, alpha_grad, beta_grad = df_d(y, yc, m, L, L_inv, K, sigma2s, alpha, beta, s_to_l=s_to_l)

    log_marginal -= KL.sum()
    alpha_grad -= dKL_da
    beta_grad -= dKL_db
    return -log_marginal, -np.r_[alpha_grad, beta_grad].reshape(-1)


def vi_comparison(
    X: np.ndarray,
    y: List[Tuple[int, float]],
    yc: List[List[Tuple[int, int]]],
    kern: GPy.kern.Kern,
    sigma2s: np.ndarray,
    alpha: np.ndarray,
    beta: np.ndarray,
    max_iters: int = 200,
    lr: float = 1e-3,
    method: str = "fr",
    optimize: str = "adam",
    get_logger: Callable = None,
) -> Tuple[Posterior, float, Dict, np.ndarray, np.ndarray]:
    """
    :param X: All locations of both direct observations and batch comparisons
    :param y: Direct observations in as a list of tuples telling location index (row in X) and observation value.
    :param yc: Batch comparisons in a list of lists of tuples. Each batch is a list and tuples tell the comparisons (winner index, loser index)
    :param kern: Prior covariance kernel
    :param sigma2s: Noise variance of the observations
    :param alpha: Initial values for alpha
    :param beta: Initial values for beta
    :param max_iter: macimum number of optimization iterations
    :param method: full rank 'fr' or mean field 'mf' methods
    :param optimize: optimization algorithm. adam or l-bfgs-B
    :param get_logger: Function for receiving the legger where the prints are forwarded.
    :return: A Tuple containing the posterior, log marginal likelihood, its gradients with respect to hyper parameters (not supported at the moment) and alpha and beta values
    """
    if method == "fr":
        recompute_posterior = recompute_posterior_fr
        s_to_l = dL_fr
    else:
        recompute_posterior = recompute_posterior_mf
        s_to_l = dL_mf

    K = kern.K(X)
    K = K + 1e-6 * np.identity(len(K))
    N = X.shape[0]

    X0 = np.r_[alpha, beta]
    args = [K, sigma2s, y, yc, recompute_posterior, s_to_l]
    if optimize == "adam":
        X, log_marginal, _ = adam(log_lik, X0.flatten(), args, bounds=None, max_it=max_iters, get_logger=get_logger)
    else:
        res = sp.optimize.minimize(fun=log_lik, x0=X0.flatten(), args=args, method="L-BFGS-B", jac=True, bounds=None)
        X = res.x.reshape(-1)
        log_marginal = res.fun
    alpha = X[: K.shape[0]].reshape(-1, 1)
    beta = X[K.shape[0] :].reshape(-1, 1)

    # Create posterior instance
    m, L, L_inv, KL, dKL_db_, dKL_da_ = recompute_posterior(alpha, beta, K)
    posterior = Posterior(mean=m, cov=L @ L.T, K=K)
    grad_dict = {}  # {'dL_dK': dF_dK - dKL_dK, 'dL_dthetaL':dL_dthetaL}
    # return posterior, log_marginal, grad_dict
    return posterior, log_marginal, grad_dict, alpha, beta



"""Base class for kernel embeddings."""

# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import List, Union

import numpy as np

from ..interfaces.standard_kernels import IStandardKernel
from ..measures import BoxDomain, GaussianMeasure, IntegrationMeasure, LebesgueMeasure
from ..typing import BoundsType


class QuadratureKernel:
    """Abstract class for a kernel augmented with integrability.

    .. note::

        Each specific implementation of this class must go with a specific standard kernel as input which
        inherits from :class:`IStandardKernel`. This is because we both want the :class:`QuadratureKernel`
        to be backend agnostic but at the same time :class:`QuadratureKernel` needs access to specifics
        of the standard kernel. For example a specific pair of :class:`QuadratureKernel` and
        :class:`IStandardKernel` is :class:`QuadratureRBF` and :class:`IRBF`. The kernel embeddings are
        implemented w.r.t. a specific integration measure, for example the :class:`LebesgueMeasure`.

    :param kern: Standard EmuKit kernel.
    :param measure: The integration measure.
    :param variable_names: The (variable) name(s) of the integral.

    """

    def __init__(
        self,
        kern: IStandardKernel,
        measure: IntegrationMeasure,
        variable_names: str,
    ) -> None:
        self.kern = kern
        self.measure = measure
        self.input_dim = self.measure.input_dim

        # Even though the domain of the integration might be unbounded, some subroutines such as the acquisition
        # optimizer require a bounded box. Hence, below we define a reasonable box for input values of the integrand
        # that can be used by such methods. The box is defined by the integration measure used (see docs there).
        # Note that the reasonable box only affects methods that use this box, e.g. the acqusition optimizers.
        # The kernel embeddings still use the originally defined domain which may be unbounded.
        self.reasonable_box = BoxDomain(name=variable_names, bounds=measure.reasonable_box())

    def K(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        """The kernel k(x1, x2) evaluated at x1 and x2.

        :param x1: First argument of the kernel.
        :param x2: Second argument of the kernel.
        :returns: The kernel at x1, x2.
        """
        return self.kern.K(x1, x2)

    # the following methods are integrals of a quadrature kernel
    def qK(self, x2: np.ndarray) -> np.ndarray:
        """The kernel with the first argument integrated out (kernel mean) evaluated at x2.

        :param x2: The locations where the kernel mean is evaluated, shape (n_points, input_dim).
        :returns: The kernel mean at x2, shape (1, n_points).
        """
        raise NotImplementedError

    def Kq(self, x1: np.ndarray) -> np.ndarray:
        """The kernel with the second argument integrated out (kernel mean) evaluated at x1.

        :param x1: The locations where the kernel mean is evaluated, shape (n_points, input_dim).
        :returns: The kernel mean at x1, shape (n_points, 1).
        """
        return self.qK(x1).T

    def qKq(self) -> float:
        """The kernel integrated over both arguments x1 and x2.

        :returns: Double integrated kernel.
        """
        raise NotImplementedError

    # the following methods are gradients of a quadrature kernel
    def dK_dx1(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        """The gradient of the kernel wrt x1 evaluated at pair x1, x2.

        :param x1: First argument of the kernel, shape (n_points N, input_dim).
        :param x2: Second argument of the kernel, shape (n_points M, input_dim).
        :return: The gradient at (x1, x2), shape (input_dim, N, M).
        """
        return self.kern.dK_dx1(x1, x2)

    def dK_dx2(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        """The gradient of the kernel wrt x2 evaluated at pair x1, x2.

        Note that this is equal to the transposed gradient of the kernel wrt x1 evaluated
        at x2 and x1 (swapped arguments).

        :param x1: First argument of the kernel, shape (n_points N, N, input_dim).
        :param x2: Second argument of the kernel, shape (n_points N, M, input_dim).
        :return: The gradient at (x1, x2), shape (input_dim, N, M).
        """
        return np.transpose(self.dK_dx1(x1=x2, x2=x1), (0, 2, 1))

    def dKdiag_dx(self, x: np.ndarray) -> np.ndarray:
        """The gradient of the diagonal of the kernel :math:`v(x):=k(x, x)` evaluated at x.

        :param x: The locations where the gradient is evaluated, shape = (n_points M, input_dim).
        :return: The gradient at x, shape (input_dim, M).
        """
        return self.kern.dKdiag_dx(x)

    def dqK_dx(self, x2: np.ndarray) -> np.ndarray:
        """The gradient of the kernel mean (integrated in first argument) evaluated at x2.

        :param x2: The locations where the gradient is evaluated, shape (n_points N, N, input_dim).
        :return: The gradient at x2, shape (input_dim, N).
        """
        raise NotImplementedError

    def dKq_dx(self, x1: np.ndarray) -> np.ndarray:
        """The gradient of the kernel mean (integrated in second argument) evaluated at x1.

        :param x1: The locations where the gradient is evaluated, shape (n_points N, N, input_dim).
        :return: The gradient with shape (N, input_dim).
        """
        return self.dqK_dx(x1).T


class QuadratureProductKernel(QuadratureKernel):
    """Abstract class for a product kernel augmented with integrability.

    The product kernel is of the form :math:`k(x, x') = \sigma^2 \prod_{i=1}^d k_i(x, x')`
    where :math:`k_i(x, x')` is a univariate kernel acting on dimension :math:`i`.

    :param kern: Standard EmuKit kernel (must be a product kernel).
    :param measure: The integration measure.
    :param variable_names: The (variable) name(s) of the integral.

    """

    def __init__(
        self,
        kern: IStandardKernel,
        measure: IntegrationMeasure,
        variable_names: str,
    ) -> None:

        super().__init__(kern=kern, measure=measure, variable_names=variable_names)

    def _qK_unscaled(self, x2: np.ndarray, skip: List[int] = None) -> np.ndarray:
        if skip is None:
            skip = []

        qK = np.ones(x2.shape[0])
        for dim in range(x2.shape[1]):
            if dim in skip:
                continue
            qK *= self._qK_1d(x2[:, dim], **self._get_univariate_parameters(dim))
        return qK[None, :]

    def qK(self, x2: np.ndarray, skip: List[int] = None) -> np.ndarray:
        return self._scale(self._qK_unscaled(x2, skip))

    def qKq(self) -> float:
        qKq = 1.0
        for dim in range(self.input_dim):
            qKq *= self._qKq_1d(**self._get_univariate_parameters(dim))
        return self._scale(qKq)

    def dqK_dx(self, x2: np.ndarray) -> np.ndarray:
        input_dim = x2.shape[1]
        dqK_dx = np.zeros([input_dim, x2.shape[0]])
        for dim in range(input_dim):
            grad_term = self._dqK_dx_1d(x2[:, dim], **self._get_univariate_parameters(dim))
            dqK_dx[dim, :] = grad_term * self._qK_unscaled(x2, skip=[dim])[0, :]
        return self._scale(dqK_dx)

    def _scale(self, z: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        """Scales the input ``z`` with a scalar value specific to the kernel.

        :param z: The value to be scaled.
        :return: The scaled value.
        """
        raise NotImplementedError

    def _get_univariate_parameters(self, dim: int) -> dict:
        """Keywords arguments used by methods related to the univariate, unscaled version
         of the kernel of dimension ``dim``.

        :param dim: The dimension.
        :return: The parameters of dimension ``dim``.
        """
        raise NotImplementedError

    # methods related to the univariate version of the kernel
    def _qK_1d(self, x: np.ndarray, **parameters) -> np.ndarray:
        """Unscaled kernel mean for univariate version of kernel.

        :param x: The locations where the kernel mean is evaluated, shape (n_points, ).
        :param parameters: The parameters of the univariate kernel.
        :return: The kernel mean of the univariate kernel evaluated at x, shape (n_points, ).
        """
        raise NotImplementedError

    def _qKq_1d(self, **parameters) -> float:
        """Unscaled kernel integrated over both arguments for univariate version of kernel.

        :param parameters: The parameters of the univariate kernel.
        :returns: Double integrated univariate kernel.
        """
        raise NotImplementedError

    def _dqK_dx_1d(self, x: np.ndarray, **parameters) -> np.ndarray:
        """Unscaled gradient of univariate version of the kernel mean.

        :param x: The locations where the kernel mean is evaluated, shape (n_points, ).
        :param parameters: The parameters of the univariate kernel.
        :return: The gradient with shape (n_points, ).
        """
        raise NotImplementedError


class LebesgueEmbedding:
    """Mixin for quadrature kernel w.r.t. Lebesgue measure.

    .. seealso::
        * :class:`emukit.quadrature.measures.LebesgueMeasure`

    """

    @classmethod
    def from_integral_bounds(
        cls, kern: IStandardKernel, integral_bounds: BoundsType, normalized: bool = False, variable_names: str = ""
    ):
        """Create the quadrature kernel w.r.t. a Lebesgue measure from the integral bounds.

        :param kern: Standard EmuKit kernel.
        :param integral_bounds: The integral bounds.
                               List of :math:`d` tuples :math:`[(a_1, b_1), (a_2, b_2), \dots, (a_d, b_d)]`,
                               where :math:`d` is the dimensionality of the integral and the tuple :math:`(a_i, b_i)`
                               contains the lower and upper integration bound of dimension :math:`i`.
        :param normalized: Weather the Lebesgue measure is normalized.
        :param variable_names: The (variable) name(s) of the integral
        :return: An instance of a quadrature kernel w.r.t. Lebesgue measure.

        """
        domain = BoxDomain(bounds=integral_bounds)
        measure = LebesgueMeasure(domain=domain, normalized=normalized)
        return cls(kern, measure, variable_names)


class GaussianEmbedding:
    """Mixin for quadrature kernel w.r.t. Gaussian measure.

    .. seealso::
        * :class:`emukit.quadrature.measures.GaussianMeasure`

    """

    @classmethod
    def from_measure_params(
        cls, kern: IStandardKernel, mean: np.ndarray, variance: Union[float, np.ndarray], variable_names: str = ""
    ):
        """Create the quadrature kernel w.r.t. a Gaussian measure from the measure parameters.

        :param kern: Standard EmuKit kernel.
        :param mean: The mean of the Gaussian measure, shape (input_dim, ).
        :param variance: The variances of the Gaussian measure, shape (input_dim, ).
                         If a scalar value is given, all dimensions will have same variance.
        :param variable_names: The (variable) name(s) of the integral
        :return: An instance of a quadrature kernel w.r.t. Gaussian measure.

        """
        measure = GaussianMeasure(mean=mean, variance=variance)
        return cls(kern, measure, variable_names)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import Any, Dict

import numpy as np

from .. import ParameterSpace

_log = logging.getLogger(__name__)

Context = Dict[str, Any]


class ContextManager:
    """
    Handles the context variables in the optimizer
    """

    def __init__(self, space: ParameterSpace, context: Context):
        """
        :param space: Parameter space of the search problem.
        :param context: Dictionary of variables and their context values.
                        These values are fixed while optimization.
        """
        self.context = context
        self.space = space
        self.contextfree_space = ParameterSpace([param for param in self.space.parameters if param.name not in context])
        self.context_space = ParameterSpace([param for param in self.space.parameters if param.name in context])

        # Find indices of context and non context variables
        self.context_idxs = []
        self.context_values = []
        for context_name, context_value in context.items():
            # Find indices of variable in the input domain
            self.context_idxs += self.space.find_parameter_index_in_model(context_name)

            # Find encoded values of context variable
            param = self.space.get_parameter_by_name(context_name)
            if hasattr(param, "encoding"):
                if context_value in param.encoding.categories:
                    _log.info(f"Parameter {context_name} fixed to {context_value}")
                    self.context_values.extend(param.encoding.get_encoding(context_value))
                else:
                    raise ValueError(f"Context value {context_value} not found in encoding for {context_name}")
            else:
                if param.check_in_domain(context_value):
                    _log.info(f"Parameter {context_name} fixed to {context_value}")
                else:
                    _log.warning(f"{context_name} with value {context_value} is out of the domain")
                self.context_values.append(context_value)

        all_idxs = list(range(space.dimensionality))
        self.non_context_idxs = [idx for idx in all_idxs if idx not in self.context_idxs]

    def expand_vector(self, x: np.ndarray) -> np.ndarray:
        """
        Expand context free parameter vector by values of the context.

        :param x: Context free parameter values as 2d-array
        :return: Parameter values with inserted context values
        """
        if len(self.context_space.parameters) == 0:
            return x
        else:
            x = np.atleast_2d(x)
            x_expanded = np.zeros((x.shape[0], self.space.dimensionality))
            x_expanded[:, np.array(self.non_context_idxs).astype(int)] = x
            x_expanded[:, np.array(self.context_idxs).astype(int)] = self.context_values
            return x_expanded
